# Copyright (c) 2024 Sean Vig
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

from libqtile import bar
from libqtile.config import Screen
from libqtile.confreader import Config
from libqtile.widget.base import _Widget


class BadWidget(_Widget):
    def __init__(self, **config):
        _Widget.__init__(self, bar.CALCULATED, **config)

    def _configure(self, qtile, bar):
        _Widget._configure(self, qtile, bar)
        # Crash!
        1 / 0


class CrashConfig(Config):
    screens = [Screen(top=bar.Bar([BadWidget(), BadWidget(hide_crash=True)], 20))]


crash_config = pytest.mark.parametrize("manager", [CrashConfig], indirect=True)


@crash_config
def test_hide_crashed_widget(manager):
    widgets = manager.c.bar["top"].items("widget")[1]
    # There should only be one widget in the bar
    assert len(widgets) == 1

    # That widget should be a ConfigErrorWidget
    assert widgets[0] == "configerrorwidget"



# Copyright (c) 2014 Rock Neurotiko
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from math import log

import psutil

from libqtile.log_utils import logger
from libqtile.widget import base


class Net(base.ThreadPoolText):
    """
    Displays interface down and up speed

    The following fields are available in the `format` string:

    - ``interface``: name of the interface
    - ``down``: download speed
    - ``down_suffix``: suffix for the download speed
    - ``down_cumulative``: cumulative download traffic
    - ``down_cumulative_suffix``: suffix for the cumulative download traffic
    - ``up``: upload speed
    - ``up_suffix``: suffix for the upload speed
    - ``up_cumulative``: cumulative upload traffic
    - ``up_cumulative_suffix``: suffix for the cumulative upload traffic
    - ``total``: total speed
    - ``total_suffix``: suffix for the total speed
    - ``total_cumulative``: cumulative total traffic
    - ``total_cumulative_suffix``: suffix for the cumulative total traffic

    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/
    """

    defaults = [
        (
            "format",
            "{interface}: {down:6.2f}{down_suffix:<2}\u2193\u2191{up:6.2f}{up_suffix:<2}",
            "Display format of down/upload/total speed of given interfaces",
        ),
        (
            "interface",
            None,
            "List of interfaces or single NIC as string to monitor, \
            None to display all active NICs combined",
        ),
        ("update_interval", 1, "The update interval."),
        ("use_bits", False, "Use bits instead of bytes per second?"),
        ("prefix", None, "Use a specific prefix for the unit of the speed."),
        (
            "cumulative_prefix",
            None,
            "Use a specific prefix for the unit of the cumulative traffic.",
        ),
    ]

    def __init__(self, **config):
        base.ThreadPoolText.__init__(self, "", **config)
        self.add_defaults(Net.defaults)

        self.factor = 1000.0
        self.allowed_prefixes = ["", "k", "M", "G", "T", "P", "E", "Z", "Y"]

        if self.use_bits:
            self.base_unit = "b"
            self.byte_multiplier = 8
        else:
            self.base_unit = "B"
            self.byte_multiplier = 1

        self.units = list(map(lambda p: p + self.base_unit, self.allowed_prefixes))

        if not isinstance(self.interface, list):
            if self.interface is None:
                self.interface = ["all"]
            elif isinstance(self.interface, str):
                self.interface = [self.interface]
            else:
                raise AttributeError(
                    f"Invalid Argument passed: {self.interface}\nAllowed Types: list, str, None"
                )
        self.stats = self.get_stats()

    def convert_b(self, num_bytes: float, prefix: str | None = None) -> tuple[float, str]:
        """Converts the number of bytes to the correct unit"""

        num_bytes *= self.byte_multiplier

        if prefix is None:
            if num_bytes > 0:
                power = int(log(num_bytes) / log(self.factor))
                power = min(power, len(self.units) - 1)
            else:
                power = 0
        else:
            power = self.allowed_prefixes.index(prefix)

        converted_bytes = num_bytes / self.factor**power
        unit = self.units[power]

        return converted_bytes, unit

    def get_stats(self):
        interfaces = {}
        if self.interface == ["all"]:
            net = psutil.net_io_counters(pernic=False)
            interfaces["all"] = {
                "down": net.bytes_recv,
                "up": net.bytes_sent,
                "total": net.bytes_recv + net.bytes_sent,
            }
            return interfaces
        else:
            net = psutil.net_io_counters(pernic=True)
            for iface in net:
                down = net[iface].bytes_recv
                up = net[iface].bytes_sent
                interfaces[iface] = {
                    "down": down,
                    "up": up,
                    "total": down + up,
                }
            return interfaces

    def poll(self):
        ret_stat = []
        try:
            new_stats = self.get_stats()
            for intf in self.interface:
                down = new_stats[intf]["down"] - self.stats[intf]["down"]
                up = new_stats[intf]["up"] - self.stats[intf]["up"]
                total = new_stats[intf]["total"] - self.stats[intf]["total"]

                down = down / self.update_interval
                up = up / self.update_interval
                total = total / self.update_interval
                down, down_suffix = self.convert_b(down, self.prefix)
                down_cumulative, down_cumulative_suffix = self.convert_b(
                    new_stats[intf]["down"], self.cumulative_prefix
                )
                up, up_suffix = self.convert_b(up, self.prefix)
                up_cumulative, up_cumulative_suffix = self.convert_b(
                    new_stats[intf]["up"], self.cumulative_prefix
                )
                total, total_suffix = self.convert_b(total, self.prefix)
                total_cumulative, total_cumulative_suffix = self.convert_b(
                    new_stats[intf]["total"], self.cumulative_prefix
                )
                self.stats[intf] = new_stats[intf]
                ret_stat.append(
                    self.format.format(
                        interface=intf,
                        down=down,
                        down_suffix=down_suffix,
                        down_cumulative=down_cumulative,
                        down_cumulative_suffix=down_cumulative_suffix,
                        up=up,
                        up_suffix=up_suffix,
                        up_cumulative=up_cumulative,
                        up_cumulative_suffix=up_cumulative_suffix,
                        total=total,
                        total_suffix=total_suffix,
                        total_cumulative=total_cumulative,
                        total_cumulative_suffix=total_cumulative_suffix,
                    )
                )

            return " ".join(ret_stat)
        except Exception:
            logger.exception("Net widget errored while polling:")



# Copyright (c) 2022 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys
from types import ModuleType

import pytest

from libqtile.bar import Bar
from libqtile.config import Screen
from libqtile.widget.launchbar import LaunchBar
from test.helpers import BareConfig


class LaunchBarTestWidget(LaunchBar):
    def get_icon_in_position(self, x, y):
        index = LaunchBar.get_icon_in_position(self, x, y)
        if index is not None:
            self.clicked_icon = self.progs[index]["name"]
        else:
            self.clicked_icon = "ERROR"
        return index


class MockXDG(ModuleType):
    def getIconPath(*args, **kwargs):  # noqa: N802
        pass


@pytest.fixture
def position(request):
    return getattr(request, "param", "top")


@pytest.fixture
def progs(request):
    print(getattr(request, "param", ""))
    return getattr(request, "param", [("test", "test", "")])


def set_progs(progs):
    return pytest.mark.parametrize("progs", progs, indirect=True)


horizontal_and_vertical = pytest.mark.parametrize("position", ["top", "left"], indirect=True)


@pytest.fixture
def launchbar_manager(request, manager_nospawn, position, progs):
    config = getattr(request, "param", dict())

    class LaunchBarConfig(BareConfig):
        screens = [
            Screen(
                **{
                    position: Bar(
                        [LaunchBarTestWidget(progs=progs, name="launchbar", padding=0, **config)],
                        28,
                    )
                }
            )
        ]

    manager_nospawn.start(LaunchBarConfig)
    yield manager_nospawn


def test_deprecated_configuration(caplog, monkeypatch):
    monkeypatch.setitem(sys.modules, "xdg.IconTheme", MockXDG("xdg.IconTheme"))
    _ = LaunchBar([("thunderbird", "thunderbird -safe-mode", "launch thunderbird in safe mode")])
    records = [r for r in caplog.records if r.msg.startswith("The use of")]
    assert records
    assert "The use of a positional argument in LaunchBar is deprecated." in records[0].msg


@horizontal_and_vertical
def test_tasklist_defaults(launchbar_manager):
    widget = launchbar_manager.c.widget["launchbar"]
    assert widget.info()["length"] > 0


@pytest.mark.parametrize(
    "position,coords,clicked",
    [
        ("top", (10, 0, 1), "one"),
        ("top", (30, 0, 1), "two"),
        ("left", (0, 30, 1), "one"),
        ("left", (0, 10, 1), "two"),
        ("right", (0, 10, 1), "one"),
        ("right", (0, 30, 1), "two"),
    ],
    indirect=["position"],
)
@set_progs([[("one", "qshell:None", ""), ("two", "qshell:None", "")]])
def test_launchbar_click(launchbar_manager, position, coords, clicked):
    def assert_clicked():
        _, value = launchbar_manager.c.widget["launchbar"].eval("self.clicked_icon")
        assert value == clicked

    launchbar_manager.c.bar[position].fake_button_press(*coords)
    assert_clicked()



# Copyright (c) 2023 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile.config
from libqtile import bar, layout
from libqtile.config import Screen
from libqtile.confreader import Config
from libqtile.widget.tasklist import TaskList
from test.layouts.layout_utils import assert_focused
from test.test_scratchpad import is_spawned, spawn_cmd


class TaskListTestWidget(TaskList):
    def __init__(self, *args, **kwargs):
        TaskList.__init__(self, *args, **kwargs)
        self._text = ""

    def calc_box_widths(self):
        ret_val = TaskList.calc_box_widths(self)
        self._text = "|".join(self.get_taskname(w) for w in self.windows)
        return ret_val

    def info(self):
        info = TaskList.info(self)
        info["text"] = self._text
        return info


@pytest.fixture
def override_xdg(request):
    return getattr(request, "param", False)


@pytest.fixture
def position(request):
    return getattr(request, "param", "top")


xdg = pytest.mark.parametrize("override_xdg", [True], indirect=True)
no_xdg = pytest.mark.parametrize("override_xdg", [False], indirect=True)
horizontal_and_vertical = pytest.mark.parametrize("position", ["top", "left"], indirect=True)


@pytest.fixture
def tasklist_manager(request, manager_nospawn, override_xdg, monkeypatch, position):
    monkeypatch.setattr("libqtile.widget.tasklist.has_xdg", override_xdg)

    config = getattr(request, "param", dict())

    class TasklistConfig(Config):
        auto_fullscreen = True
        groups = [
            libqtile.config.ScratchPad(
                "SCRATCHPAD",
                dropdowns=[
                    libqtile.config.DropDown("dd-a", spawn_cmd("dd-a"), on_focus_lost_hide=False),
                ],
            ),
            libqtile.config.Group("a"),
            libqtile.config.Group("b"),
        ]
        layouts = [layout.Stack()]
        floating_layout = libqtile.resources.default_config.floating_layout
        keys = []
        mouse = []
        screens = [
            Screen(**{position: bar.Bar([TaskListTestWidget(name="tasklist", **config)], 28)})
        ]

    manager_nospawn.start(TasklistConfig)
    yield manager_nospawn


def configure_tasklist(**config):
    """Decorator to pass configuration to widget."""
    return pytest.mark.parametrize("tasklist_manager", [config], indirect=True)


@horizontal_and_vertical
def test_tasklist_defaults(tasklist_manager):
    widget = tasklist_manager.c.widget["tasklist"]

    tasklist_manager.test_window("One")
    tasklist_manager.test_window("Two")
    assert widget.info()["text"] == "One|Two"

    # Test floating
    tasklist_manager.c.window.toggle_floating()
    assert widget.info()["text"] == "One|V Two"

    tasklist_manager.c.window.toggle_floating()
    assert widget.info()["text"] == "One|Two"

    # Test maximize
    tasklist_manager.c.window.toggle_maximize()
    assert widget.info()["text"] == "One|[] Two"

    tasklist_manager.c.window.toggle_maximize()
    assert widget.info()["text"] == "One|Two"

    # Test minimize
    tasklist_manager.c.window.toggle_minimize()
    assert widget.info()["text"] == "One|_ Two"

    tasklist_manager.c.window.toggle_minimize()
    assert widget.info()["text"] == "One|Two"


def test_tasklist_skip_taskbar_defaults(tasklist_manager):
    widget = tasklist_manager.c.widget["tasklist"]
    tasklist_manager.c.group["SCRATCHPAD"].dropdown_reconfigure("dd-a")

    tasklist_manager.test_window("one")
    assert_focused(tasklist_manager, "one")

    # dd-a has no window associated yet
    assert "window" not in tasklist_manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")

    # First toggling: wait for window
    tasklist_manager.c.group["SCRATCHPAD"].dropdown_toggle("dd-a")
    is_spawned(tasklist_manager, "dd-a")
    assert_focused(tasklist_manager, "dd-a")
    assert (
        tasklist_manager.c.group["SCRATCHPAD"].dropdown_info("dd-a")["window"]["name"] == "dd-a"
    )

    if tasklist_manager.c.core.info()["backend"] == "x11":
        # check that window's _NET_WM_STATE contains _NET_WM_STATE_SKIP_TASKBAR
        net_wm_state = tasklist_manager.c.window.eval(
            'self.window.get_property("_NET_WM_STATE", "ATOM", unpack=int)'
        )[1]
        skip_taskbar = tasklist_manager.c.window.eval(
            'self.qtile.core.conn.atoms["_NET_WM_STATE_SKIP_TASKBAR"]'
        )[1]
        assert skip_taskbar in net_wm_state
        assert tasklist_manager.c.window.eval("self.window.get_wm_type()")[1] == "normal"
        assert widget.info()["text"] == "one"


@configure_tasklist(txt_minimized="(min) ", txt_maximized="(max) ", txt_floating="(float) ")
def test_tasklist_custom_text(tasklist_manager):
    widget = tasklist_manager.c.widget["tasklist"]

    tasklist_manager.test_window("One")
    tasklist_manager.test_window("Two")
    assert widget.info()["text"] == "One|Two"

    # Test floating
    tasklist_manager.c.window.toggle_floating()
    assert widget.info()["text"] == "One|(float) Two"

    tasklist_manager.c.window.toggle_floating()
    assert widget.info()["text"] == "One|Two"

    # Test maximize
    tasklist_manager.c.window.toggle_maximize()
    assert widget.info()["text"] == "One|(max) Two"

    tasklist_manager.c.window.toggle_maximize()
    assert widget.info()["text"] == "One|Two"

    # Test minimize
    tasklist_manager.c.window.toggle_minimize()
    assert widget.info()["text"] == "One|(min) Two"

    tasklist_manager.c.window.toggle_minimize()
    assert widget.info()["text"] == "One|Two"


@configure_tasklist(markup_minimized="_{}_", markup_maximized="[{}]", markup_floating="V{}V")
def test_tasklist_custom_markup(tasklist_manager):
    """markup_* options override txt_*"""
    widget = tasklist_manager.c.widget["tasklist"]

    tasklist_manager.test_window("One")
    tasklist_manager.test_window("Two")
    assert widget.info()["text"] == "One|Two"

    # Test floating
    tasklist_manager.c.window.toggle_floating()
    assert widget.info()["text"] == "One|VTwoV"

    tasklist_manager.c.window.toggle_floating()
    assert widget.info()["text"] == "One|Two"

    # Test maximize
    tasklist_manager.c.window.toggle_maximize()
    assert widget.info()["text"] == "One|[Two]"

    tasklist_manager.c.window.toggle_maximize()
    assert widget.info()["text"] == "One|Two"

    # Test minimize
    tasklist_manager.c.window.toggle_minimize()
    assert widget.info()["text"] == "One|_Two_"

    tasklist_manager.c.window.toggle_minimize()
    assert widget.info()["text"] == "One|Two"


@configure_tasklist(markup_focused="({})", markup_focused_floating="[{}]")
def test_tasklist_focused_and_floating(tasklist_manager):
    widget = tasklist_manager.c.widget["tasklist"]

    tasklist_manager.test_window("One")
    tasklist_manager.test_window("Two")
    assert widget.info()["text"] == "One|(Two)"

    # Test floating
    tasklist_manager.c.window.toggle_floating()
    assert widget.info()["text"] == "One|[Two]"

    tasklist_manager.c.window.toggle_floating()
    assert widget.info()["text"] == "One|(Two)"


@pytest.mark.parametrize(
    "position,coords",
    [("top", (0, 0, 1)), ("right", (0, 0, 1)), ("left", (0, 599, 1))],
    indirect=["position"],
)
@configure_tasklist(margin=0)
def test_tasklist_click_task(tasklist_manager, position, coords):
    tasklist_manager.test_window("One")
    tasklist_manager.test_window("Two")

    # Current focused window is "Two"
    assert tasklist_manager.c.window.info()["name"] == "Two"

    # Click in top left corner of bar means we click on "One"
    # which should focus the window
    # margin is set to 0 as value set by widget_defaults means text would otherwise
    # mean text does not start at x=0
    tasklist_manager.c.bar[position].fake_button_press(*coords)
    assert tasklist_manager.c.window.info()["name"] == "One"


@xdg
@configure_tasklist(theme_mode="non-existent-mode")
@pytest.mark.xfail
def test_tasklist_bad_theme_mode(tasklist_manager):
    msgs = tasklist_manager.get_log_buffer()
    assert "Unexpected theme_mode (non-existent-mode). Theme icons will be disabled." in msgs


@no_xdg
@configure_tasklist(theme_mode="non-existent-mode")
@pytest.mark.xfail
def test_tasklist_no_xdg(tasklist_manager):
    msgs = tasklist_manager.get_log_buffer()
    assert "You must install pyxdg to use theme icons." in msgs


@horizontal_and_vertical
@configure_tasklist(stretch=False)
def test_tasklist_no_stretch(tasklist_manager, position):
    widget = tasklist_manager.c.widget["tasklist"]
    tasklist_manager.test_window("One")
    width_one = widget.info()["length"]

    tasklist_manager.test_window("Two")
    width_two = widget.info()["length"]

    assert width_one != width_two



# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

import libqtile.bar
import libqtile.config
import libqtile.confreader
import libqtile.layout
from libqtile.widget import CurrentLayout


def get_widget_config(widget, config):
    config.screens = [
        libqtile.config.Screen(top=libqtile.bar.Bar([widget], 10)),
    ]
    config.layouts = [
        libqtile.layout.Columns(),
        libqtile.layout.Max(),
        libqtile.layout.Stack(),
    ]
    return config


def test_current_layout(manager_nospawn, minimal_conf_noscreen):
    config = get_widget_config(CurrentLayout(), minimal_conf_noscreen)
    manager_nospawn.start(config)
    widget = manager_nospawn.c.widget["currentlayout"]

    layout = widget.info()["text"]
    assert layout == "columns"

    manager_nospawn.c.next_layout()
    layout = widget.info()["text"]
    assert layout == "max"

    manager_nospawn.c.prev_layout()
    layout = widget.info()["text"]
    assert layout == "columns"

    widget.bar.fake_button_press(0, 0, button=1)
    layout = widget.info()["text"]
    assert layout == "max"

    widget.bar.fake_button_press(0, 0, button=2)
    layout = widget.info()["text"]
    assert layout == "columns"

    manager_nospawn.c.screen.next_group()
    manager_nospawn.c.to_layout_index(-1)
    layout = widget.info()["text"]
    assert layout == "stack"

    manager_nospawn.c.screen.prev_group()
    layout = widget.info()["text"]
    assert layout == "columns"


def test_current_layout_icon_mode(manager_nospawn, minimal_conf_noscreen):
    config = get_widget_config(CurrentLayout(mode="icon"), minimal_conf_noscreen)
    manager_nospawn.start(config)
    widget = manager_nospawn.c.widget["currentlayout"]
    img_length = int(widget.eval("self.img_length")[1])
    padding = int(widget.eval("self.actual_padding")[1])

    length = int(widget.eval("self.length")[1])
    assert length == img_length + padding * 2

    widget.bar.fake_button_press(0, 0, button=3)
    length = int(widget.eval("self.length")[1])
    assert length != img_length + padding * 2

    widget.bar.fake_button_press(0, 0, button=3)
    length = int(widget.eval("self.length")[1])
    assert length == img_length + padding * 2


def test_current_layout_text_mode(manager_nospawn, minimal_conf_noscreen):
    config = get_widget_config(CurrentLayout(mode="text"), minimal_conf_noscreen)
    manager_nospawn.start(config)
    widget = manager_nospawn.c.widget["currentlayout"]
    text_length = int(widget.eval("super(type(self), self).calculate_length()")[1])

    length = int(widget.eval("self.length")[1])
    assert length == text_length

    widget.bar.fake_button_press(0, 0, button=3)
    length = int(widget.eval("self.length")[1])
    assert length != text_length

    widget.bar.fake_button_press(0, 0, button=3)
    length = int(widget.eval("self.length")[1])
    assert length == text_length


def test_current_layout_both_mode(manager_nospawn, minimal_conf_noscreen):
    config = get_widget_config(CurrentLayout(mode="both"), minimal_conf_noscreen)
    manager_nospawn.start(config)
    widget = manager_nospawn.c.widget["currentlayout"]
    img_length = int(widget.eval("self.img_length")[1])
    padding = int(widget.eval("self.actual_padding")[1])
    text_length = int(widget.eval("super(type(self), self).calculate_length()")[1])

    length = int(widget.eval("self.length")[1])
    assert length == text_length + img_length + padding

    widget.bar.fake_button_press(0, 0, button=3)
    length = int(widget.eval("self.length")[1])
    assert length == text_length + img_length + padding



# Copyright (c) 2024 Thomas Krug
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import multiprocessing
import os
import subprocess
import tempfile
import time

import test

repo_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../"))
socket_path = tempfile.NamedTemporaryFile().name


def run_qtile(backend):
    cmd = os.path.join(repo_path, "bin/qtile")
    args = [cmd, "start", "-s", socket_path]
    args.extend(["-b", backend.name])

    env = os.environ.copy()
    if backend.name == "wayland":
        env.update(backend.env)

    proc = subprocess.Popen(args, env=env, stdout=subprocess.PIPE)
    out, err = proc.communicate(timeout=10)
    exitcode = proc.returncode
    return exitcode


def stop_qtile(code):
    cmd = os.path.join(repo_path, "bin/qtile")
    args = [cmd, "cmd-obj", "-o", "cmd", "-s", socket_path, "-f", "shutdown"]
    if code:
        args.extend(["-a", str(code)])
    proc = subprocess.Popen(args, stdout=subprocess.PIPE)
    proc.communicate(timeout=10)
    exitcode = proc.returncode
    return exitcode


def deferred_stop(code=0):
    # wait for qtile process to start
    wait = 10
    while not test.helpers.can_connect_qtile(socket_path):
        time.sleep(1)
        if not wait:
            raise Exception("timeout waiting for qtile process")
        wait -= 1

    stop_qtile(code)


# in these testcases there are two qtile instances
# one started by the helpers.TestManager,
# which is used to evaluate code coverage
# and a second instance started by run_qtile,
# which is used to test the actual exit code behavior
def test_exitcode_default(backend):
    proc = multiprocessing.Process(target=deferred_stop)
    proc.start()

    exitcode = run_qtile(backend)
    proc.join()
    assert exitcode == 0


def test_exitcode_explicit(backend):
    code = 23

    proc = multiprocessing.Process(target=deferred_stop, args=(code,))
    proc.start()

    exitcode = run_qtile(backend)
    proc.join()
    assert exitcode == code



# Copyright (c) 2025 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio
import contextlib

import pytest

from libqtile.interactive.repl import (
    REPL_PORT,
    TERMINATOR,
    QtileREPLServer,
    get_completions,
)


def test_get_completions_top_level():
    local_vars = {"qtile": "dummy", "qtiles": 123}
    result = get_completions("qti", local_vars)
    assert "qtile" in result
    assert "qtiles" in result


def test_get_completions_attribute():
    class Dummy:
        def method(self):
            pass

        val = 42

    local_vars = {"dummy": Dummy()}
    result = get_completions("dummy.me", local_vars)
    assert "dummy.method(" in result

    result = get_completions("dummy.va", local_vars)
    assert "dummy.val" in result


def test_get_completions_invalid_expr():
    result = get_completions("invalid..expr", {})
    assert result == []


@pytest.mark.anyio
async def test_repl_server_executes_code():
    repl = QtileREPLServer()
    locals_dict = {"x": 123}

    # Start the REPL server in a background task
    start_task = asyncio.create_task(repl.start(locals_dict=locals_dict))

    # Wait for the server to bind the port
    await asyncio.sleep(0.1)

    reader, writer = await asyncio.open_connection("localhost", REPL_PORT)

    try:
        # Read welcome message
        welcome = await reader.read(4096)
        assert b"Connected to Qtile REPL" in welcome

        # Send a simple expression to evaluate
        writer.write(b"x\n" + f"{TERMINATOR}\n".encode())
        await writer.drain()

        # Read REPL result
        result = await reader.readuntil(f"{TERMINATOR}\n".encode())
        assert "123" in result.decode()

    finally:
        writer.close()
        await writer.wait_closed()

        # Stop the REPL server
        await repl.stop()

        # Cancel the server task
        start_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await start_task



# Copyright (c) 2024 Thomas Krug
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time

import pytest

import libqtile
from test.helpers import Retry


class DGroupsConfig(libqtile.confreader.Config):
    auto_fullscreen = True
    groups = [libqtile.config.Group("a"), libqtile.config.Group("b")]
    layouts = [libqtile.layout.MonadTall()]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []


class DGroupsSpawnConfig(DGroupsConfig):
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b", spawn=["xterm"]),
    ]


dgroups_config = pytest.mark.parametrize("manager", [DGroupsConfig], indirect=True)
dgroups_spawn_config = pytest.mark.parametrize("manager", [DGroupsSpawnConfig], indirect=True)


@dgroups_config
def test_dgroup_persist(manager):
    # create dgroup
    gname = "c"
    manager.c.addgroup(gname, persist=True)

    # switch to dgroup
    manager.c.group[gname].toscreen()

    # start window
    one = manager.test_window("test1")

    # close window
    manager.kill_window(one)

    # wait for window to close and group to NOT be destroyed
    time.sleep(2)

    # check if dgroup still exists
    assert len(manager.c.get_groups()) == 3


@dgroups_config
def test_dgroup_nonpersist(manager):
    # create dgroup
    gname = "c"
    manager.c.addgroup(gname)

    # switch to dgroup
    manager.c.group[gname].toscreen()

    # start window
    one = manager.test_window("test1")

    # close window
    manager.kill_window(one)

    # wait for window to close and group to be destroyed
    time.sleep(2)

    # check if dgroup does not exist anymore
    assert len(manager.c.get_groups()) == 2


@dgroups_spawn_config
def test_dgroup_spawn_in_group(manager):
    @Retry(ignore_exceptions=(AssertionError,))
    def wait_for_window():
        assert len(manager.c.windows()) > 0

    wait_for_window()
    assert not manager.c.group["a"].info()["windows"]
    assert manager.c.group["b"].info()["windows"]



# Copyright (c) 2024 Saath Satheeshkumar (saths008)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from libqtile.config import Bar, Screen
from libqtile.confreader import Config
from libqtile.widget import redshift


def mock_run(argv, check):
    pass


@pytest.fixture(scope="function")
def patched_redshift(monkeypatch):
    class PatchedRedshift(redshift.Redshift):
        def __init__(self, **config):
            monkeypatch.setattr("subprocess.run", mock_run)
            redshift.Redshift.__init__(self, **config)
            self.name = "redshift"

    yield PatchedRedshift


@pytest.fixture(scope="function")
def redshift_manager(manager_nospawn, request, patched_redshift):
    if manager_nospawn.backend.name == "wayland":
        pytest.skip("Skipping test on Wayland.")

    class GroupConfig(Config):
        screens = [
            Screen(
                top=Bar(
                    [
                        patched_redshift(
                            update_interval=10,
                            **getattr(request, "param", dict()),
                        )
                    ],
                    30,
                )
            )
        ]

    manager_nospawn.start(GroupConfig)

    yield manager_nospawn


def config(**kwargs):
    return pytest.mark.parametrize("redshift_manager", [kwargs], indirect=True)


def test_defaults(redshift_manager):
    widget = redshift_manager.c.widget["redshift"]

    def text():
        return widget.info()["text"]

    def click():
        redshift_manager.c.bar["top"].fake_button_press(0, 0, 1)

    disabled_txt = "󱠃"
    enabled_txt = "󰛨"
    assert text() == disabled_txt

    scroll_vals = [
        "Brightness: 1.0",
        "Temperature: 1700",
        "Gamma: 1.0:1.0:1.0",
        enabled_txt,
    ]

    click()

    for _, val in enumerate(scroll_vals):
        widget.scroll_up()
        assert text() == val

    click()
    for _ in range(len(scroll_vals)):
        widget.scroll_down()
        # Test that scroll only works when
        # widget is enabled
        assert text() == disabled_txt


@config(disabled_txt="Redshift disabled", enabled_txt="Redshift enabled")
def test_changed_default_txt_non_fmted(redshift_manager):
    widget = redshift_manager.c.widget["redshift"]

    def text():
        return widget.info()["text"]

    def click():
        redshift_manager.c.bar["top"].fake_button_press(0, 0, 1)

    disabled_txt = "Redshift disabled"
    enabled_txt = "Redshift enabled"

    assert text() == disabled_txt

    click()
    assert text() == enabled_txt

    scroll_vals = [
        "Brightness: 1.0",
        "Temperature: 1700",
        "Gamma: 1.0:1.0:1.0",
        enabled_txt,
    ]

    for _, val in enumerate(scroll_vals):
        widget.scroll_up()
        assert text() == val

    # move enabled_txt to the first index
    scroll_vals.remove(enabled_txt)
    scroll_vals = [enabled_txt] + scroll_vals

    for _, val in enumerate(reversed(scroll_vals)):
        widget.scroll_down()
        assert text() == val


@config(
    brightness=0.4,
    disabled_txt="brightness{brightness}, temp{temperature}, r{gamma_red}, g{gamma_green}, b{gamma_blue}, is_enabled{is_enabled}",
    enabled_txt="brightness{brightness}, temp{temperature}, r{gamma_red}, g{gamma_green}, b{gamma_blue}, is_enabled{is_enabled}",
    gamma_red=0.2,
    gamma_green=0.2,
    gamma_blue=0.2,
    temperature=1200,
)
def test_changed_default_txt_fmted(redshift_manager):
    widget = redshift_manager.c.widget["redshift"]

    def text():
        return widget.info()["text"]

    def click():
        redshift_manager.c.bar["top"].fake_button_press(0, 0, 1)

    brightness = 0.4
    gamma_red = 0.2
    gamma_green = 0.2
    gamma_blue = 0.2
    gamma_val = redshift.GammaGroup(gamma_red, gamma_green, gamma_blue)
    temp = 1200

    comm_str = f"brightness{brightness}, temp{temp}, r{gamma_red}, g{gamma_green}, b{gamma_blue}"
    disabled_txt = comm_str + f", is_enabled{False}"
    enabled_txt = comm_str + f", is_enabled{True}"

    assert text() == disabled_txt

    click()
    assert text() == enabled_txt

    scroll_vals = [
        f"Brightness: {brightness}",
        f"Temperature: {temp}",
        f"Gamma: {gamma_val._redshift_fmt()}",
        enabled_txt,
    ]

    for _, val in enumerate(scroll_vals):
        widget.scroll_up()
        assert text() == val

    scroll_vals.remove(enabled_txt)
    scroll_vals = [enabled_txt] + scroll_vals

    for _, val in enumerate(reversed(scroll_vals)):
        widget.scroll_down()
        assert text() == val


@config(
    disabled_txt="Disabled",
    enabled_txt="Enabled t:{temperature} b:{brightness}",
    brightness=0.5,
    brightness_step=0.2,
    temperature=1200,
    temperature_step=101,
)
def test_increase_decrease_temp_brightness(redshift_manager):
    widget = redshift_manager.c.widget["redshift"]

    def text():
        return widget.info()["text"]

    def click():
        redshift_manager.c.bar["top"].fake_button_press(0, 0, 1)

    def right_click():
        redshift_manager.c.bar["top"].fake_button_press(0, 0, 3)

    disabled_txt = "Disabled"
    enabled_txt = "Enabled t:1200 b:0.5"

    assert text() == disabled_txt

    click()

    assert text() == enabled_txt

    widget.scroll_up()

    # Test that the right click respects brightness boundaries
    brightness_vals = [0.5, 0.7, 0.9, 1.0, 1.0, 1.0]
    for _, val in enumerate(brightness_vals):
        assert text() == f"Brightness: {val}"
        right_click()

    # Test that the left click respects brightness boundaries
    brightness_vals = [1.0, 0.8, 0.6, 0.4, 0.2, 0.1, 0.1]
    for _, val in enumerate(brightness_vals):
        assert text() == f"Brightness: {val}"
        click()

    widget.scroll_down()  # Enabled

    assert text() == "Enabled t:1200 b:0.1"

    widget.scroll_down()  # Gamma
    widget.scroll_down()  # Temperature

    # Test that the right click respects temperature boundaries
    temp_vals = [1200]

    temp_lower_bound = 1000
    temp_upper_bound = 25000
    while temp_vals[-1] < temp_upper_bound:
        new_temp_val = temp_vals[-1] + 101
        new_temp_val = min(new_temp_val, temp_upper_bound)
        temp_vals.append(new_temp_val)

    for _, val in enumerate(temp_vals):
        assert text() == f"Temperature: {val}"
        right_click()

    # Test that the left click respects temperature boundaries
    temp_vals = [temp_upper_bound]
    while temp_vals[-1] > temp_lower_bound:
        new_temp_val = temp_vals[-1] - 101
        new_temp_val = max(new_temp_val, temp_lower_bound)
        temp_vals.append(new_temp_val)

    for _, val in enumerate(temp_vals):
        assert text() == f"Temperature: {val}"
        click()

    widget.scroll_up()  # Gamma
    widget.scroll_up()  # Temperature
    assert text() == "Enabled t:1000 b:0.1"



# Copyright (c) 2025 e-devnull
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from libqtile.widget import NetUP
from test.widgets.conftest import FakeBar


def test_host_is_empty():
    netup = NetUP()
    assert netup.poll() == "N/A"


def test_invalid_method():
    netup = NetUP(host="localhost", method="icmp")
    assert netup.poll() == "N/A"


def test_invalid_port():
    netup = NetUP(host="localhost", method="tcp", port="port")
    assert netup.poll() == "N/A"


def test_ping_success(monkeypatch, fake_qtile, fake_window):
    def mock_ping(*args, **kwargs):
        class MockResult:
            def __init__(self, returncode):
                self.returncode = returncode

        return MockResult(returncode=0)

    monkeypatch.setattr("libqtile.widget.netup.run", mock_ping)
    netup = NetUP(host="localhost", method="ping")
    fakebar = FakeBar([netup], window=fake_window)
    netup._configure(fake_qtile, fakebar)

    assert netup.poll() == "NET " + netup.up_string
    assert netup.layout.colour == netup.up_foreground


def test_ping_fail(monkeypatch, fake_qtile, fake_window):
    def mock_ping(*args, **kwargs):
        class MockResult:
            def __init__(self, returncode):
                self.returncode = returncode

        return MockResult(returncode=1)

    monkeypatch.setattr("libqtile.widget.netup.run", mock_ping)
    netup = NetUP(host="localhost", method="ping")
    fakebar = FakeBar([netup], window=fake_window)
    netup._configure(fake_qtile, fakebar)

    assert netup.poll() == "NET " + netup.down_string
    assert netup.layout.colour == netup.down_foreground


def test_tcp_success(monkeypatch, fake_qtile, fake_window):
    def mock_check_tcp(*args, **kwargs):
        return 0

    monkeypatch.setattr(NetUP, "check_tcp", mock_check_tcp)
    netup = NetUP(host="localhost", method="tcp", port=443)
    fakebar = FakeBar([netup], window=fake_window)
    netup._configure(fake_qtile, fakebar)

    assert netup.poll() == "NET " + netup.up_string
    assert netup.layout.colour == netup.up_foreground


def test_tcp_fail(monkeypatch, fake_qtile, fake_window):
    def mock_check_tcp(*args, **kwargs):
        return -1

    monkeypatch.setattr(NetUP, "check_tcp", mock_check_tcp)
    netup = NetUP(host="localhost", method="tcp", port=443)
    fakebar = FakeBar([netup], window=fake_window)
    netup._configure(fake_qtile, fakebar)

    assert netup.poll() == "NET " + netup.down_string
    assert netup.layout.colour == netup.down_foreground



#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import subprocess


def run_qtile(args):
    cmd = os.path.join(os.path.dirname(__file__), "..", "bin", "qtile")
    argv = [cmd]
    argv.extend(args)
    proc = subprocess.Popen(argv, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    assert proc.returncode == 0
    stdout = stdout.decode()
    stderr = stderr.decode()
    return (stdout, stderr)


def test_cmd_help_subcommand():
    args = ["help"]
    stdout, stderr = run_qtile(args)
    assert "usage: qtile" in stdout
    assert stderr == ""


def test_cmd_help_param():
    args = ["--help"]
    stdout, stderr = run_qtile(args)
    assert "usage: qtile" in stdout
    assert stderr == ""



# Copyright (c) 2025 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

from libqtile.scripts.repl import TERMINATOR, is_code_complete, read_full_response


class MockSocket:
    def recv(self, *args, **kwargs):
        return (f"Qtile test\nREPL client\n{TERMINATOR}\n").encode()


def test_read_full_response_basic():
    # Create a mock socket with recv method
    sock = MockSocket()

    response = read_full_response(sock)
    assert "Qtile test" in response
    assert "REPL client" in response
    # The terminator should be stripped off
    assert TERMINATOR not in response


@pytest.mark.parametrize(
    "input,valid",
    [
        ("x+1", True),
        ("def a():", False),
        ("def a():\n  print(1)", False),
        ("def a():\n  print(1)\n", True),
        ("def a():\n  print(1)\ndef b():\n  print(2)\n", True),
        ("while True:\n", False),
    ],
    ids=[
        "SingleExpression",
        "IncompleteFunction",
        "MultilineIncomplete",
        "MultilineComplete",
        "MultiMultiline",
        "IncompleteContext",
    ],
)
def test_is_code_complete(input, valid):
    assert is_code_complete(f"{input}\n") is valid



# Copyright (c) 2009 Aldo Cortesi
# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2011 Anshuman Bhaduri
# Copyright (c) 2012 Tycho Andersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio
from multiprocessing import Value

import pytest

import libqtile.log_utils
import libqtile.utils
from libqtile import config, hook, layout
from libqtile.config import Match
from libqtile.resources import default_config
from test.conftest import BareConfig, dualmonitor
from test.helpers import Retry


class Call:
    def __init__(self, val):
        self.val = val

    def __call__(self, val):
        self.val = val


class NoArgCall(Call):
    def __call__(self):
        self.val += 1


@pytest.fixture
def hook_fixture():
    libqtile.log_utils.init_log()

    yield

    hook.clear()


def test_cannot_fire_unknown_event():
    with pytest.raises(libqtile.utils.QtileError):
        hook.fire("unknown")


@pytest.mark.usefixtures("hook_fixture")
def test_hook_calls_subscriber():
    test = Call(0)
    hook.subscribe.group_window_add(test)
    hook.fire("group_window_add", 8)
    assert test.val == 8


@pytest.mark.usefixtures("hook_fixture")
def test_hook_calls_subscriber_async():
    val = 0

    async def co(new_val):
        nonlocal val
        val = new_val

    hook.subscribe.group_window_add(co)
    hook.fire("group_window_add", 8)

    assert val == 8


@pytest.mark.usefixtures("hook_fixture")
def test_hook_calls_subscriber_async_co():
    val = 0

    async def co(new_val):
        nonlocal val
        val = new_val

    hook.subscribe.group_window_add(co(8))
    hook.fire("group_window_add")

    assert val == 8


@pytest.mark.usefixtures("hook_fixture")
def test_hook_calls_subscriber_async_in_existing_loop():
    async def t():
        val = 0

        async def co(new_val):
            nonlocal val
            val = new_val

        hook.subscribe.group_window_add(co(8))
        hook.fire("group_window_add")
        await asyncio.sleep(0)
        assert val == 8

    asyncio.run(t())


@pytest.mark.usefixtures("hook_fixture")
def test_subscribers_can_be_added_removed():
    test = Call(0)
    hook.subscribe.group_window_add(test)
    assert hook.subscriptions
    hook.clear()
    assert not hook.subscriptions


@pytest.mark.usefixtures("hook_fixture")
def test_can_unsubscribe_from_hook():
    test = Call(0)

    hook.subscribe.group_window_add(test)
    hook.fire("group_window_add", 3)
    assert test.val == 3

    hook.unsubscribe.group_window_add(test)
    hook.fire("group_window_add", 4)
    assert test.val == 3


def test_can_subscribe_to_startup_hooks(manager_nospawn):
    config = BareConfig
    for attr in dir(default_config):
        if not hasattr(config, attr):
            setattr(config, attr, getattr(default_config, attr))
    manager = manager_nospawn

    manager.startup_once_calls = Value("i", 0)
    manager.startup_calls = Value("i", 0)
    manager.startup_complete_calls = Value("i", 0)

    def inc_startup_once_calls():
        manager.startup_once_calls.value += 1

    def inc_startup_calls():
        manager.startup_calls.value += 1

    def inc_startup_complete_calls():
        manager.startup_complete_calls.value += 1

    hook.subscribe.startup_once(inc_startup_once_calls)
    hook.subscribe.startup(inc_startup_calls)
    hook.subscribe.startup_complete(inc_startup_complete_calls)

    manager.start(config)
    assert manager.startup_once_calls.value == 1
    assert manager.startup_calls.value == 1
    assert manager.startup_complete_calls.value == 1

    # Restart and check that startup_once doesn't fire again
    manager.terminate()
    manager.start(config, no_spawn=True)
    assert manager.startup_once_calls.value == 1
    assert manager.startup_calls.value == 2
    assert manager.startup_complete_calls.value == 2


@pytest.mark.usefixtures("hook_fixture")
def test_can_update_by_selection_change(manager):
    test = Call(0)
    hook.subscribe.selection_change(test)
    hook.fire("selection_change", "hello")
    assert test.val == "hello"


@pytest.mark.usefixtures("hook_fixture")
def test_can_call_by_selection_notify(manager):
    test = Call(0)
    hook.subscribe.selection_notify(test)
    hook.fire("selection_notify", "hello")
    assert test.val == "hello"


@pytest.mark.usefixtures("hook_fixture")
def test_resume_hook(manager):
    test = NoArgCall(0)
    hook.subscribe.resume(test)
    hook.fire("resume")
    assert test.val == 1


@pytest.mark.usefixtures("hook_fixture")
def test_suspend_hook(manager):
    test = NoArgCall(0)
    hook.subscribe.suspend(test)
    hook.fire("suspend")
    assert test.val == 1


@pytest.mark.usefixtures("hook_fixture")
def test_custom_hook_registry():
    """Tests ability to create custom hook registries"""
    test = NoArgCall(0)

    custom = hook.Registry("test")
    custom.register_hook(hook.Hook("test_hook"))
    custom.subscribe.test_hook(test)

    assert test.val == 0

    # Test ability to fire third party hooks
    custom.fire("test_hook")
    assert test.val == 1

    # Check core hooks are not included in custom registry
    with pytest.raises(libqtile.utils.QtileError):
        custom.fire("client_managed")

    # Check custom hooks are not in core registry
    with pytest.raises(libqtile.utils.QtileError):
        hook.fire("test_hook")


@pytest.mark.usefixtures("hook_fixture")
def test_user_hook(manager_nospawn):
    config = BareConfig
    for attr in dir(default_config):
        if not hasattr(config, attr):
            setattr(config, attr, getattr(default_config, attr))
    manager = manager_nospawn

    manager.custom_no_arg_text = Value("u", "A")
    manager.custom_text = Value("u", "A")

    # Define two functions: first takes no args, second takes a single arg
    def predefined_text():
        with manager.custom_no_arg_text.get_lock():
            manager.custom_no_arg_text.value = "B"

    def defined_text(text):
        with manager.custom_text.get_lock():
            manager.custom_text.value = text

    hook.subscribe.user("set_text")(predefined_text)
    hook.subscribe.user("define_text")(defined_text)

    # Check values are as initialised
    manager.start(config)
    assert manager.custom_no_arg_text.value == "A"
    assert manager.custom_text.value == "A"

    # Check hooked function with no args
    manager.c.fire_user_hook("set_text")
    assert manager.custom_no_arg_text.value == "B"

    # Check hooked function with a single arg
    manager.c.fire_user_hook("define_text", "C")
    assert manager.custom_text.value == "C"


def test_shutdown(manager_nospawn):
    def inc_shutdown_calls():
        manager_nospawn.shutdown_calls.value += 1

    manager_nospawn.shutdown_calls = Value("i", 0)
    hook.subscribe.shutdown(inc_shutdown_calls)

    manager_nospawn.start(BareConfig)
    manager_nospawn.c.shutdown()
    assert manager_nospawn.shutdown_calls.value == 1


@dualmonitor
def test_setgroup(manager_nospawn):
    @Retry(ignore_exceptions=(AssertionError))
    def assert_inc_calls(num: int):
        assert manager_nospawn.setgroup_calls.value == num

    def inc_setgroup_calls():
        manager_nospawn.setgroup_calls.value += 1

    manager_nospawn.setgroup_calls = Value("i", 0)
    hook.subscribe.setgroup(inc_setgroup_calls)

    # Starts with two because of the dual screen
    manager_nospawn.start(BareConfig)
    assert_inc_calls(2)

    manager_nospawn.c.switch_groups("a", "b")
    assert_inc_calls(3)

    manager_nospawn.c.to_screen(1)
    assert_inc_calls(4)
    manager_nospawn.c.to_screen(1)
    assert_inc_calls(4)

    manager_nospawn.c.next_screen()
    assert_inc_calls(5)

    manager_nospawn.c.prev_screen()
    assert_inc_calls(6)

    manager_nospawn.c.group.switch_groups("b")
    assert_inc_calls(7)


class CallGroupname:
    def __init__(self):
        self.groupname = ""

    def __call__(self, groupname):
        self.groupname = groupname


@Retry(ignore_exceptions=(AssertionError))
def assert_groupname(mgr_nospawn, groupname):
    _, _groupname = mgr_nospawn.c.eval("self.config.test.groupname")
    assert _groupname == groupname


@pytest.mark.usefixtures("hook_fixture")
def test_addgroup(manager_nospawn):
    class AddgroupConfig(BareConfig):
        test = CallGroupname()
        hook.subscribe.addgroup(test)

    manager_nospawn.start(AddgroupConfig)
    assert_groupname(manager_nospawn, "d")
    manager_nospawn.c.addgroup("e")
    assert_groupname(manager_nospawn, "e")


@pytest.mark.usefixtures("hook_fixture")
def test_delgroup(manager_nospawn):
    class DelgroupConfig(BareConfig):
        test = CallGroupname()
        hook.subscribe.delgroup(test)

    manager_nospawn.start(DelgroupConfig)
    manager_nospawn.c.delgroup("e")
    assert_groupname(manager_nospawn, "")
    manager_nospawn.c.delgroup("d")
    assert_groupname(manager_nospawn, "d")


def test_changegroup(manager_nospawn):
    @Retry(ignore_exceptions=(AssertionError))
    def assert_inc_calls(num: int):
        assert manager_nospawn.changegroup_calls.value == num

    def inc_changegroup_calls():
        manager_nospawn.changegroup_calls.value += 1

    manager_nospawn.changegroup_calls = Value("i", 0)
    hook.subscribe.changegroup(inc_changegroup_calls)

    # Starts with four beacuase of four groups in BareConfig
    manager_nospawn.start(BareConfig)
    assert_inc_calls(4)

    manager_nospawn.c.group.set_label("Test")
    assert_inc_calls(5)

    manager_nospawn.c.addgroup("e")
    assert_inc_calls(6)
    manager_nospawn.c.addgroup("e")
    assert_inc_calls(6)

    manager_nospawn.c.delgroup("e")
    assert_inc_calls(7)
    manager_nospawn.c.delgroup("e")
    assert_inc_calls(7)


def test_focus_change(manager_nospawn):
    @Retry(ignore_exceptions=(AssertionError))
    def assert_inc_calls(num: int):
        assert manager_nospawn.focus_change_calls.value == num

    def inc_focus_change_calls():
        manager_nospawn.focus_change_calls.value += 1

    manager_nospawn.focus_change_calls = Value("i", 0)
    hook.subscribe.focus_change(inc_focus_change_calls)

    manager_nospawn.start(BareConfig)
    assert_inc_calls(1)

    manager_nospawn.test_window("Test Window")
    assert_inc_calls(2)

    manager_nospawn.c.group.focus_by_index(0)
    assert_inc_calls(3)
    manager_nospawn.c.group.focus_by_index(1)
    assert_inc_calls(3)

    manager_nospawn.test_window("Test Focus Change")
    assert_inc_calls(4)

    manager_nospawn.c.group.focus_back()
    assert_inc_calls(5)

    manager_nospawn.c.group.focus_by_name("Test Focus Change")
    assert_inc_calls(6)
    manager_nospawn.c.group.focus_by_name("Test Focus")
    assert_inc_calls(6)

    manager_nospawn.c.group.next_window()
    assert_inc_calls(7)

    manager_nospawn.c.group.prev_window()
    assert_inc_calls(8)

    manager_nospawn.c.window.kill()
    assert_inc_calls(9)


def test_float_change(manager_nospawn):
    @Retry(ignore_exceptions=(AssertionError))
    def assert_inc_calls(num: int):
        assert manager_nospawn.float_change_calls.value == num

    def inc_float_change_calls():
        manager_nospawn.float_change_calls.value += 1

    manager_nospawn.float_change_calls = Value("i", 0)
    hook.subscribe.float_change(inc_float_change_calls)

    manager_nospawn.start(BareConfig)
    manager_nospawn.test_window("Test Window")

    manager_nospawn.c.window.enable_floating()
    assert_inc_calls(1)
    manager_nospawn.c.window.enable_floating()
    assert_inc_calls(1)

    manager_nospawn.c.window.disable_floating()
    assert_inc_calls(2)
    manager_nospawn.c.window.disable_floating()
    assert_inc_calls(2)

    manager_nospawn.c.window.toggle_floating()
    assert_inc_calls(3)

    manager_nospawn.c.window.toggle_floating()
    manager_nospawn.c.window.move_floating(0, 0)
    assert_inc_calls(5)

    manager_nospawn.c.window.toggle_floating()
    manager_nospawn.c.window.resize_floating(10, 10)
    assert_inc_calls(7)

    manager_nospawn.c.window.toggle_floating()
    manager_nospawn.c.window.set_position_floating(0, 0)
    assert_inc_calls(9)

    manager_nospawn.c.window.toggle_floating()
    manager_nospawn.c.window.set_size_floating(100, 100)
    assert_inc_calls(11)


class CallGroupWindow:
    def __init__(self):
        self.window = ""
        self.group = ""

    def __call__(self, group, win):
        self.group = group.name
        self.window = win.name


@Retry(ignore_exceptions=(AssertionError))
def assert_group_window(mgr_nospawn, group, window):
    _, _group = mgr_nospawn.c.eval("self.config.test.group")
    _, _window = mgr_nospawn.c.eval("self.config.test.window")
    assert _group == group
    assert _window == window


@pytest.mark.usefixtures("hook_fixture")
def test_group_window_add(manager_nospawn):
    class AddGroupWindowConfig(BareConfig):
        test = CallGroupWindow()
        hook.subscribe.group_window_add(test)

    manager_nospawn.start(AddGroupWindowConfig)
    manager_nospawn.test_window("Test Window")
    assert_group_window(manager_nospawn, "a", "Test Window")


@pytest.mark.usefixtures("hook_fixture")
def test_group_window_remove(manager_nospawn):
    class RemoveGroupWindowConfig(BareConfig):
        test = CallGroupWindow()
        hook.subscribe.group_window_remove(test)

    manager_nospawn.start(RemoveGroupWindowConfig)
    manager_nospawn.test_window("Test Window")
    manager_nospawn.c.window.kill()
    assert_group_window(manager_nospawn, "a", "Test Window")


class CallWindow:
    def __init__(self):
        self.window = ""

    def __call__(self, window):
        self.window = window.name


@Retry(ignore_exceptions=(AssertionError))
def assert_window(mgr_nospawn, window):
    _, _window = mgr_nospawn.c.eval("self.config.test.window")
    assert _window == window


@pytest.mark.usefixtures("hook_fixture")
def test_client_new(manager_nospawn):
    class ClientNewConfig(BareConfig):
        test = CallWindow()
        hook.subscribe.client_new(test)

    manager_nospawn.start(ClientNewConfig)
    manager_nospawn.test_window("Test Client")
    assert_window(manager_nospawn, "Test Client")


@pytest.mark.usefixtures("hook_fixture")
def test_client_managed(manager_nospawn):
    class ClientManagedConfig(BareConfig):
        test = CallWindow()
        hook.subscribe.client_managed(test)

    manager_nospawn.start(ClientManagedConfig)
    manager_nospawn.test_window("Test Client")
    assert_window(manager_nospawn, "Test Client")

    manager_nospawn.test_window("Test Static")
    manager_nospawn.c.group.focus_back()
    manager_nospawn.c.window.static()
    assert_window(manager_nospawn, "Test Client")


@pytest.mark.usefixtures("hook_fixture")
def test_client_killed(manager_nospawn):
    class ClientKilledConfig(BareConfig):
        test = CallWindow()
        hook.subscribe.client_killed(test)

    manager_nospawn.start(ClientKilledConfig)
    manager_nospawn.test_window("Test Client")
    manager_nospawn.c.window.kill()
    assert_window(manager_nospawn, "Test Client")


@pytest.mark.usefixtures("hook_fixture")
def test_client_focus(manager_nospawn):
    class ClientFocusConfig(BareConfig):
        test = CallWindow()
        hook.subscribe.client_focus(test)

    manager_nospawn.start(ClientFocusConfig)
    manager_nospawn.test_window("Test Client")
    assert_window(manager_nospawn, "Test Client")

    manager_nospawn.test_window("Test Focus")
    manager_nospawn.c.group.focus_back()
    assert_window(manager_nospawn, "Test Client")


@pytest.mark.usefixtures("hook_fixture")
def test_client_mouse_enter(manager_nospawn):
    class ClientMouseEnterConfig(BareConfig):
        test = CallWindow()
        hook.subscribe.client_mouse_enter(test)

    manager_nospawn.start(ClientMouseEnterConfig)
    manager_nospawn.test_window("Test Client")
    manager_nospawn.backend.fake_click(0, 0)
    assert_window(manager_nospawn, "Test Client")


@pytest.mark.usefixtures("hook_fixture")
def test_client_name_updated(manager_nospawn):
    class ClientNameUpdatedConfig(BareConfig):
        test = CallWindow()
        hook.subscribe.client_name_updated(test)

    manager_nospawn.start(ClientNameUpdatedConfig)
    manager_nospawn.test_window("Test Client", new_title="Test NameUpdated")
    assert_window(manager_nospawn, "Test NameUpdated")


@pytest.mark.usefixtures("hook_fixture")
def test_client_urgent_hint_changed(manager_nospawn, backend_name):
    if backend_name == "wayland":
        pytest.skip("Core not listening to XDG request_activate_event ?")

    class ClientUrgentHintChangedConfig(BareConfig):
        groups = [
            config.Group("a"),
            config.Group("b", matches=[Match(title="Test Client")]),
        ]
        focus_on_window_activation = "urgent"
        test = CallWindow()
        hook.subscribe.client_urgent_hint_changed(test)

    manager_nospawn.start(ClientUrgentHintChangedConfig)
    manager_nospawn.test_window("Test Client", urgent_hint=True)
    assert_window(manager_nospawn, "Test Client")


class CallLayoutGroup:
    def __init__(self):
        self.layout = ""
        self.group = ""

    def __call__(self, layout, group):
        self.layout = layout.name
        self.group = group.name


@Retry(ignore_exceptions=(AssertionError))
def assert_layout_group(mgr_nospawn, layout, group):
    _, _layout = mgr_nospawn.c.eval("self.config.test.layout")
    assert _layout == layout
    _, _group = mgr_nospawn.c.eval("self.config.test.group")
    assert _group == group


@pytest.mark.usefixtures("hook_fixture")
def test_layout_change(manager_nospawn):
    class ClientLayoutChange(BareConfig):
        layouts = [layout.stack.Stack(), layout.columns.Columns()]
        test = CallLayoutGroup()
        hook.subscribe.layout_change(test)

    manager_nospawn.start(ClientLayoutChange)
    assert_layout_group(manager_nospawn, "stack", "a")

    manager_nospawn.c.group.setlayout("columns")
    assert_layout_group(manager_nospawn, "columns", "a")

    manager_nospawn.c.screen.next_group()
    assert_layout_group(manager_nospawn, "stack", "b")

    manager_nospawn.c.screen.prev_group()
    assert_layout_group(manager_nospawn, "columns", "a")

    manager_nospawn.c.screen.toggle_group()
    assert_layout_group(manager_nospawn, "stack", "b")

    manager_nospawn.c.next_layout()
    assert_layout_group(manager_nospawn, "columns", "b")

    manager_nospawn.c.prev_layout()
    assert_layout_group(manager_nospawn, "stack", "b")


@pytest.mark.usefixtures("hook_fixture")
def test_net_wm_icon_change(manager_nospawn, backend_name):
    if backend_name == "wayland":
        pytest.skip("X11 only.")

    class ClientNewConfig(BareConfig):
        test = CallWindow()
        hook.subscribe.net_wm_icon_change(test)

    manager_nospawn.start(ClientNewConfig)
    manager_nospawn.test_window("Test Client")
    assert_window(manager_nospawn, "Test Client")


@pytest.mark.usefixtures("hook_fixture")
def test_screen_change(manager_nospawn):
    @Retry(ignore_exceptions=(AssertionError))
    def assert_inc_calls(num: int):
        assert manager_nospawn.screen_change_calls.value == num

    def inc_screen_change_calls(event):
        manager_nospawn.screen_change_calls.value += 1

    manager_nospawn.screen_change_calls = Value("i", 0)
    hook.subscribe.screen_change(inc_screen_change_calls)

    manager_nospawn.start(BareConfig)
    assert_inc_calls(1)


@pytest.mark.usefixtures("hook_fixture")
def test_screens_reconfigured(manager_nospawn):
    @Retry(ignore_exceptions=(AssertionError))
    def assert_inc_calls(num: int):
        assert manager_nospawn.screens_reconfigured_calls.value == num

    def inc_screens_reconfigured_calls():
        manager_nospawn.screens_reconfigured_calls.value += 1

    manager_nospawn.screens_reconfigured_calls = Value("i", 0)
    hook.subscribe.screens_reconfigured(inc_screens_reconfigured_calls)

    manager_nospawn.start(BareConfig)
    manager_nospawn.c.reconfigure_screens()
    assert_inc_calls(1)


@dualmonitor
@pytest.mark.usefixtures("hook_fixture")
def test_current_screen_change(manager_nospawn):
    @Retry(ignore_exceptions=(AssertionError))
    def assert_inc_calls(num: int):
        assert manager_nospawn.current_screen_change_calls.value == num

    def inc_current_screen_change_calls():
        manager_nospawn.current_screen_change_calls.value += 1

    manager_nospawn.current_screen_change_calls = Value("i", 0)
    hook.subscribe.current_screen_change(inc_current_screen_change_calls)

    manager_nospawn.start(BareConfig)

    manager_nospawn.c.to_screen(1)
    assert_inc_calls(1)
    manager_nospawn.c.to_screen(1)
    assert_inc_calls(1)

    manager_nospawn.c.next_screen()
    assert_inc_calls(2)

    manager_nospawn.c.prev_screen()
    assert_inc_calls(3)



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012, 2014 Tycho Andersen
# Copyright (c) 2013 Craig Barnes
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio

import pytest

import libqtile.bar
import libqtile.config
import libqtile.confreader
import libqtile.layout
import libqtile.log_utils
import libqtile.widget
from libqtile.command.base import CommandObject, expose_command
from libqtile.command.client import CommandClient
from libqtile.command.interface import CommandError, IPCCommandInterface
from libqtile.confreader import Config
from libqtile.ipc import Client, IPCError
from libqtile.lazy import lazy
from test.conftest import dualmonitor
from test.helpers import Retry


class CallConfig(Config):
    keys = [
        libqtile.config.Key(
            ["control"],
            "j",
            lazy.layout.down(),
        ),
        libqtile.config.Key(
            ["control"],
            "k",
            lazy.layout.up(),
        ),
    ]
    mouse = []
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
    ]
    layouts = [
        libqtile.layout.Stack(num_stacks=1),
        libqtile.layout.Max(),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    screens = [
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.GroupBox(),
                    libqtile.widget.TextBox(),
                ],
                20,
            ),
        )
    ]
    auto_fullscreen = True


call_config = pytest.mark.parametrize("manager", [CallConfig], indirect=True)


@call_config
def test_layout_filter(manager):
    manager.test_window("one")
    manager.test_window("two")
    assert manager.c.get_groups()["a"]["focus"] == "two"
    manager.c.simulate_keypress(["control"], "j")
    assert manager.c.get_groups()["a"]["focus"] == "one"
    manager.c.simulate_keypress(["control"], "k")
    assert manager.c.get_groups()["a"]["focus"] == "two"


@call_config
def test_param_hoisting(manager):
    manager.test_window("two")

    client = Client(manager.sockfile)
    command = IPCCommandInterface(client)
    cmd_client = CommandClient(command)

    # 'zomg' is not a valid warp command
    with pytest.raises(IPCError):
        cmd_client.navigate("window", None).call("focus", warp="zomg", lifted=True)

    cmd_client.navigate("window", None).call("focus", warp=False, lifted=True)

    # 'zomg' is not a valid bar position
    with pytest.raises(IPCError):
        cmd_client.call("hide_show_bar", position="zomg", lifted=True)

    cmd_client.call("hide_show_bar", position="top", lifted=True)

    # 'zomg' is not a valid font size
    with pytest.raises(IPCError):
        cmd_client.navigate("widget", "textbox").call("set_font", fontsize="zomg", lifted=True)

    cmd_client.navigate("widget", "textbox").call("set_font", fontsize=12, lifted=True)


class FakeCommandObject(CommandObject):
    @staticmethod
    @expose_command()
    def one():
        pass

    @expose_command()
    def one_self(self):
        pass

    @expose_command()
    def two(self, a):
        pass

    @expose_command()
    def three(self, a, b=99):
        pass

    def _items(self, name):
        return None

    def _select(self, name, sel):
        return None


def test_doc():
    c = FakeCommandObject()
    assert "one()" in c.doc("one")
    assert "one_self()" in c.doc("one_self")
    assert "two(a)" in c.doc("two")
    assert "three(a, b=99)" in c.doc("three")


def test_commands():
    c = FakeCommandObject()
    assert len(c.commands()) == 9


def test_command():
    c = FakeCommandObject()
    assert c.command("one")
    assert not c.command("nonexistent")


class DecoratedTextBox(libqtile.widget.TextBox):
    @expose_command("mapped")
    def exposed(self):
        return "OK"


class ServerConfig(Config):
    auto_fullscreen = True
    keys = []
    mouse = []
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
    ]
    layouts = [
        libqtile.layout.Stack(num_stacks=1),
        libqtile.layout.Stack(num_stacks=2),
        libqtile.layout.Stack(num_stacks=3),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    screens = [
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.TextBox(name="one"),
                ],
                20,
            ),
        ),
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    DecoratedTextBox(name="two"),
                ],
                20,
            ),
        ),
    ]


server_config = pytest.mark.parametrize("manager", [ServerConfig], indirect=True)


@server_config
def test_cmd_commands(manager):
    assert manager.c.commands()
    assert manager.c.layout.commands()
    assert manager.c.screen.bar["bottom"].commands()


@server_config
def test_cmd_eval_namespace(manager):
    assert manager.c.eval("__name__") == (True, "libqtile.core.manager")


@server_config
def test_call_unknown(manager):
    with pytest.raises(libqtile.command.client.SelectError, match="Not valid child or command"):
        manager.c.nonexistent

    manager.c.layout
    with pytest.raises(libqtile.command.client.SelectError, match="Not valid child or command"):
        manager.c.layout.nonexistent


@dualmonitor
@server_config
def test_items_qtile(manager):
    v = manager.c.items("group")
    assert v[0]
    assert sorted(v[1]) == ["a", "b", "c"]

    assert manager.c.items("layout") == (True, [0, 1, 2])

    v = manager.c.items("widget")
    assert not v[0]
    assert sorted(v[1]) == ["one", "two"]

    assert manager.c.items("bar") == (False, ["bottom"])
    t, lst = manager.c.items("window")
    assert t
    assert len(lst) == 2
    assert manager.c.window[lst[0]]
    assert manager.c.items("screen") == (True, [0, 1])


@dualmonitor
@server_config
def test_select_qtile(manager):
    assert manager.c.layout.info()["group"] == "a"
    assert len(manager.c.layout.info()["stacks"]) == 1
    assert len(manager.c.layout[2].info()["stacks"]) == 3
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        manager.c.layout[99]

    assert manager.c.group.info()["name"] == "a"
    assert manager.c.group["c"].info()["name"] == "c"
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        manager.c.group["nonexistent"]

    assert manager.c.widget["one"].info()["name"] == "one"
    with pytest.raises(CommandError, match="No object widget"):
        manager.c.widget.info()

    assert manager.c.bar["bottom"].info()["position"] == "bottom"

    manager.test_window("one")
    wid = manager.c.window.info()["id"]
    assert manager.c.window[wid].info()["id"] == wid

    assert manager.c.screen.info()["index"] == 0
    assert manager.c.screen[1].info()["index"] == 1
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        manager.c.screen[22]


@server_config
def test_items_group(manager):
    group = manager.c.group

    manager.test_window("test")
    wid = manager.c.window.info()["id"]

    assert group.items("window") == (True, [wid])
    assert group.items("layout") == (True, [0, 1, 2])
    assert group.items("screen") == (True, [])


@dualmonitor
@server_config
def test_select_group(manager):
    group = manager.c.group

    assert group.layout.info()["group"] == "a"
    assert len(group.layout.info()["stacks"]) == 1
    assert len(group.layout[2].info()["stacks"]) == 3

    with pytest.raises(CommandError):
        manager.c.group.window.info()
    manager.test_window("test")
    wid = manager.c.window.info()["id"]

    assert group.window.info()["id"] == wid
    assert group.window[wid].info()["id"] == wid
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        group.window["foo"]

    assert group.screen.info()["index"] == 0
    assert group["b"].screen.info()["index"] == 1
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        group.screen[0]


@server_config
def test_items_screen(manager):
    s = manager.c.screen
    assert s.items("layout") == (True, [0, 1, 2])

    manager.test_window("test")
    wid = manager.c.window.info()["id"]
    assert s.items("window") == (True, [wid])

    assert s.items("bar") == (False, ["bottom"])


@server_config
def test_select_screen(manager):
    screen = manager.c.screen
    assert screen.layout.info()["group"] == "a"
    assert len(screen.layout.info()["stacks"]) == 1
    assert len(screen.layout[2].info()["stacks"]) == 3

    with pytest.raises(CommandError):
        manager.c.window.info()

    manager.test_window("test")
    wid = manager.c.window.info()["id"]
    assert screen.window.info()["id"] == wid
    assert screen.window[wid].info()["id"] == wid

    with pytest.raises(CommandError, match="No object"):
        screen.bar.info()
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        screen.bar["top"]

    assert screen.bar["bottom"].info()["position"] == "bottom"


@server_config
def test_items_bar(manager):
    assert manager.c.bar["bottom"].items("screen") == (True, [])


@dualmonitor
@server_config
def test_select_bar(manager):
    assert manager.c.screen[1].bar["bottom"].screen.info()["index"] == 1
    b = manager.c.bar
    assert b["bottom"].screen.info()["index"] == 0
    with pytest.raises(CommandError):
        b.screen.info()


@server_config
def test_items_layout(manager):
    assert manager.c.layout.items("screen") == (True, [])
    assert manager.c.layout.items("group") == (True, [])


@server_config
def test_select_layout(manager):
    layout = manager.c.layout

    assert layout.screen.info()["index"] == 0
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        layout.screen[0]

    assert layout.group.info()["name"] == "a"
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        layout.group["a"]


@dualmonitor
@server_config
def test_items_window(manager):
    manager.test_window("test")
    window = manager.c.window
    window.info()["id"]

    assert window.items("group") == (True, [])
    assert window.items("layout") == (True, [0, 1, 2])
    assert window.items("screen") == (True, [])


@dualmonitor
@server_config
def test_select_window(manager):
    manager.test_window("test")
    window = manager.c.window
    window.info()["id"]

    assert window.group.info()["name"] == "a"
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        window.group["a"]

    assert len(window.layout.info()["stacks"]) == 1
    assert len(window.layout[1].info()["stacks"]) == 2

    assert window.screen.info()["index"] == 0
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        window.screen[0]


@server_config
def test_items_widget(manager):
    assert manager.c.widget["one"].items("bar") == (True, [])


@server_config
def test_select_widget(manager):
    widget = manager.c.widget["one"]
    assert widget.bar.info()["position"] == "bottom"
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        widget.bar["bottom"]


def test_core_node(manager, backend_name):
    assert manager.c.core.info()["backend"] == backend_name


def test_lazy_arguments(manager_nospawn):
    # Decorated function to be bound to key presses
    @lazy.function
    def test_func(qtile, value, multiplier=1):
        qtile.test_func_output = value * multiplier

    config = ServerConfig
    config.keys = [
        libqtile.config.Key(
            ["control"],
            "j",
            test_func(10),
        ),
        libqtile.config.Key(["control"], "k", test_func(5, multiplier=100)),
    ]

    manager_nospawn.start(config)

    manager_nospawn.c.simulate_keypress(["control"], "j")
    _, val = manager_nospawn.c.eval("self.test_func_output")
    assert val == "10"

    manager_nospawn.c.simulate_keypress(["control"], "k")
    _, val = manager_nospawn.c.eval("self.test_func_output")
    assert val == "500"


def test_lazy_function_coroutine(manager_nospawn):
    """Test that lazy.function accepts coroutines."""

    @Retry(ignore_exceptions=(AssertionError,))
    def assert_func_text(manager, value):
        _, text = manager.c.eval("self.test_func_output")
        assert text == value

    @lazy.function
    async def test_async_func(qtile, value):
        await asyncio.sleep(0.1)
        qtile.test_func_output = value

    config = ServerConfig
    config.keys = [libqtile.config.Key(["control"], "k", test_async_func("qtile"))]

    manager_nospawn.start(config)

    manager_nospawn.c.simulate_keypress(["control"], "k")
    assert_func_text(manager_nospawn, "qtile")


def test_decorators_direct_call():
    widget = DecoratedTextBox()
    undecorated = libqtile.widget.TextBox()

    cmds = ["exposed", "mapped"]

    for cmd in cmds:
        # Check new command is exposed
        assert cmd in widget.commands()
        # Check commands are not in widget with same parent class
        assert cmd not in undecorated.commands()

    assert widget.exposed() == "OK"
    assert widget.mapped() == "OK"


def test_decorators_deprecated_direct_call():
    widget = DecoratedTextBox()
    assert widget.cmd_exposed() == "OK"


def test_decorators_deprecated_method():
    class CmdWidget(libqtile.widget.TextBox):
        def cmd_exposed(self):
            pass

    assert "exposed" in CmdWidget().commands()


@dualmonitor
@server_config
def test_decorators_manager_call(manager):
    widget = manager.c.widget["two"]
    assert widget.exposed() == "OK"
    assert widget.mapped() == "OK"



# Copyright (c) 2025 Emma Nora Theuer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Widget specific tests

from unittest.mock import MagicMock, patch

from libqtile.widget.tuned_manager import TunedManager


def test_find_mode():
    # Mocking subprocess.run to return a specific output
    with patch("subprocess.run") as mock_run:
        mock_run.return_value.stdout = "Current active profile: balanced-battery\n"

        widget = TunedManager()
        mode = widget.find_mode()

        assert mode == "balanced-battery"
        assert mock_run.call_count == 2  # Called during init and explicitly


def test_update_bar():
    with (
        patch("subprocess.run") as mock_run,
        patch.object(TunedManager, "bar", create=True) as mock_bar,
    ):
        mock_run.return_value.stdout = "Current active profile: powersave\n"
        mock_bar.draw = MagicMock()

        widget = TunedManager()
        widget.update_bar()

        assert widget.current_mode == "powersave"
        assert widget.text == "powersave"
        mock_bar.draw.assert_called_once()


def test_next_mode():
    with patch.object(TunedManager, "execute_command") as mock_execute_command:
        widget = TunedManager()
        widget.modes = ["powersave", "balanced-battery", "throughput-performance"]
        widget.current_mode = "powersave"

        widget.next_mode()

        mock_execute_command.assert_called_once_with(1)


def test_previous_mode():
    with patch.object(TunedManager, "execute_command") as mock_execute_command:
        widget = TunedManager()
        widget.modes = ["powersave", "balanced-battery", "throughput-performance"]
        widget.current_mode = "balanced-battery"

        widget.previous_mode()

        mock_execute_command.assert_called_once_with(0)



# Copyright (c) 2023, elParaguayo. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
import subprocess
import sys
import tempfile
import textwrap
from pathlib import Path

import pytest

from libqtile.scripts.migrations._base import NoChange
from test.test_check import run_qtile_check


class SourceCode:
    def __init__(self, test):
        self.source = textwrap.dedent(test.input)
        self.expected = textwrap.dedent(test.output)
        self.dir = tempfile.TemporaryDirectory()
        self.source_path = Path(self.dir.name) / "config.py"
        self.needs_lint = not isinstance(test, NoChange)

    def __enter__(self):
        self.source_path.write_text(self.source)
        return self

    def __exit__(self, type, value, traceback):
        self.dir.cleanup()
        del self.dir


class MigrationTester:
    def __init__(self, test_id, test):
        self.test_id = test_id
        self.test = test
        self.cmd = Path(__file__).parent / ".." / ".." / "bin" / "qtile"

    def assert_migrate(self):
        if not self.test.source:
            assert False, f"{self.test_id} has no tests."
        argv = [
            sys.executable,
            self.cmd,
            "migrate",
            "--yes",
            "-r",
            self.test_id,
            "-c",
            self.test.source_path,
        ]
        try:
            subprocess.check_call(argv)
        except subprocess.CalledProcessError:
            assert False

        updated = Path(f"{self.test.source_path}").read_text()
        assert updated == self.test.expected

    def assert_lint(self):
        argv = [
            sys.executable,
            self.cmd,
            "migrate",
            "--lint",
            "-r",
            self.test_id,
            "-c",
            self.test.source_path,
        ]
        try:
            output = subprocess.run(argv, capture_output=True, check=True)
        except subprocess.CalledProcessError:
            assert False

        assert any(self.test_id in line for line in output.stdout.decode().splitlines())

    def assert_check(self):
        if not self.test.source:
            assert False, f"{self.test_id} has no Check test."

        self.assert_migrate()

        assert run_qtile_check(self.test.source_path)


@pytest.fixture
def migration_tester(request):
    test_id, test = getattr(request, "param", (None, None))
    if test is None:
        test = NoChange("")
    with SourceCode(test) as sc:
        yield MigrationTester(test_id, sc)



import contextlib
import os
import subprocess

import pytest
import xcffib
import xcffib.testing
import xcffib.xproto
import xcffib.xtest

from libqtile.backend.x11.core import Core
from libqtile.backend.x11.xcbq import Connection
from test.helpers import (
    HEIGHT,
    SECOND_HEIGHT,
    SECOND_WIDTH,
    WIDTH,
    Backend,
    BareConfig,
    Retry,
    TestManager,
)


@Retry(ignore_exceptions=(xcffib.ConnectionException,), return_on_fail=True)
def can_connect_x11(disp=":0", *, ok=None):
    if ok is not None and not ok():
        raise AssertionError()

    conn = xcffib.connect(display=disp)
    conn.disconnect()
    return True


@contextlib.contextmanager
def xvfb():
    with xcffib.testing.XvfbTest():
        display = os.environ["DISPLAY"]
        if not can_connect_x11(display):
            raise OSError("Xvfb did not come up")

        yield


@pytest.fixture(scope="session")
def display():  # noqa: F841
    with xvfb():
        yield os.environ["DISPLAY"]


def start_x11_and_poll_connection(args, display):
    proc = subprocess.Popen(args)

    if can_connect_x11(display, ok=lambda: proc.poll() is None):
        return proc

    # we weren't able to get a display up
    if proc.poll() is None:
        raise AssertionError(f"Unable to connect to running {args[0]}")
    else:
        raise AssertionError(
            f"Unable to start {args[0]}, quit with return code {proc.returncode}"
        )


def stop_x11(proc, display, display_file):
    # Kill xephyr only if it is running
    if proc is not None:
        if proc.poll() is None:
            proc.kill()
        proc.wait()

    # clean up the lock file for the display we allocated
    try:
        display_file.close()
        os.remove(xcffib.testing.lock_path(int(display[1:])))
    except OSError:
        pass


class Xephyr:
    """Spawn Xephyr instance

    Set-up a Xephyr instance with the given parameters.  The Xephyr instance
    must be started, and then stopped.
    """

    def __init__(self, outputs, xoffset=None, xtrace=False):
        self.outputs = outputs
        if xoffset is None:
            self.xoffset = WIDTH
        else:
            self.xoffset = xoffset

        self.proc = None  # Handle to Xephyr instance, subprocess.Popen object
        self.display = None
        self.xephyr_display_file = None

        self.xtrace = xtrace
        self.xtrace_proc = None
        self.xtrace_display = None
        self.xtrace_display_file = None
        self.xephyr_display = None

    def __enter__(self):
        try:
            self.start_xephyr()
        except:  # noqa: E722
            self.stop_xephyr()
            raise

        return self

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.stop_xephyr()

    def start_xephyr(self):
        """Start Xephyr instance

        Starts the Xephyr instance and sets the `self.display` to the display
        which is used to setup the instance.
        """
        # get a new display
        display, self.xephyr_display_file = xcffib.testing.find_display()
        self.display = f":{display}"
        self.xephyr_display = self.display

        # build up arguments
        args = [
            "Xephyr",
            "-name",
            "qtile_test",
            self.xephyr_display,
            "-ac",
            "-screen",
            f"{WIDTH}x{HEIGHT}",
        ]
        if self.outputs == 2:
            args.extend(
                [
                    "-origin",
                    f"{self.xoffset},0",
                    "-screen",
                    f"{SECOND_WIDTH}x{SECOND_HEIGHT}",
                ]
            )
            args.extend(["+xinerama"])
            args.extend(["-extension", "RANDR"])

        start_x11_and_poll_connection(args, self.xephyr_display)

        if self.xtrace:
            # because we run Xephyr without auth and xtrace requires auth, we
            # need to add some x11 auth here for the Xephyr display our xtrace
            # will fail:
            subprocess.check_call(["xauth", "generate", self.xephyr_display])
            display, self.xtrace_display_file = xcffib.testing.find_display()
            self.xtrace_display = f":{display}"
            self.display = self.xtrace_display
            args = [
                "xtrace",
                "--timestamps",
                "-k",
                "-d",
                self.xephyr_display,
                "-D",
                self.xtrace_display,
            ]
            start_x11_and_poll_connection(args, self.xtrace_display)

    def stop_xephyr(self):
        stop_x11(self.proc, self.xephyr_display, self.xephyr_display_file)
        if self.xtrace:
            stop_x11(self.xtrace_proc, self.xtrace_display, self.xtrace_display_file)


@contextlib.contextmanager
def x11_environment(outputs, **kwargs):
    """This backend needs a Xephyr instance running"""
    with xvfb():
        with Xephyr(outputs, **kwargs) as x:
            yield x


@pytest.fixture(scope="function")
def xmanager(request, xephyr):
    """
    This replicates the `manager` fixture except that the x11 backend is hard-coded. We
    cannot simply parametrize the `backend_name` fixture module-wide because it gets
    parametrized by `pytest_generate_tests` in test/conftest.py and only one of these
    parametrize calls can be used.
    """
    config = getattr(request, "param", BareConfig)
    backend = XBackend({"DISPLAY": xephyr.display}, args=[xephyr.display])

    with TestManager(backend, request.config.getoption("--debuglog")) as manager:
        manager.display = xephyr.display
        manager.start(config)
        yield manager


@pytest.fixture(scope="function")
def xmanager_nospawn(request, xephyr):
    """
    This replicates the `manager` fixture except that the x11 backend is hard-coded. We
    cannot simply parametrize the `backend_name` fixture module-wide because it gets
    parametrized by `pytest_generate_tests` in test/conftest.py and only one of these
    parametrize calls can be used.
    """
    backend = XBackend({"DISPLAY": xephyr.display}, args=[xephyr.display])

    with TestManager(backend, request.config.getoption("--debuglog")) as manager:
        manager.display = xephyr.display
        yield manager


@pytest.fixture(scope="function")
def conn(xmanager):
    conn = Connection(xmanager.display)
    yield conn
    conn.finalize()


class XBackend(Backend):
    name = "x11"

    def __init__(self, env, args=()):
        self.env = env
        self.args = args
        self.core = Core
        self.manager = None

    def fake_click(self, x, y):
        """Click at the specified coordinates"""
        conn = Connection(self.env["DISPLAY"])
        root = conn.default_screen.root.wid
        xtest = conn.conn(xcffib.xtest.key)
        xtest.FakeInput(6, 0, xcffib.xproto.Time.CurrentTime, root, x, y, 0)
        xtest.FakeInput(4, 1, xcffib.xproto.Time.CurrentTime, root, 0, 0, 0)
        xtest.FakeInput(5, 1, xcffib.xproto.Time.CurrentTime, root, 0, 0, 0)
        conn.conn.flush()
        self.manager.c.sync()
        conn.finalize()

    def get_all_windows(self):
        """Get a list of all windows in ascending order of Z position"""
        conn = Connection(self.env["DISPLAY"])
        root = conn.default_screen.root.wid
        q = conn.conn.core.QueryTree(root).reply()
        wins = list(q.children)
        conn.finalize()
        return wins



# Copyright (c) 2024 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import datetime
import sys
from importlib import reload

import pytest

from libqtile.config import Bar, Screen
from libqtile.widget import vertical_clock
from test.helpers import BareConfig


# Mock Datetime object that returns a set datetime and also
# has a simplified timezone method to check functionality of
# the widget.
class MockDatetime(datetime.datetime):
    @classmethod
    def now(cls, *args, **kwargs):
        return cls(2024, 1, 1, 10, 20, 30)

    def astimezone(self, tzone=None):
        if tzone is None:
            return self
        return self + tzone.utcoffset(None)


@pytest.fixture
def patched_clock(monkeypatch):
    # Stop system importing these modules in case they exist on environment
    monkeypatch.setitem(sys.modules, "pytz", None)
    monkeypatch.setitem(sys.modules, "dateutil", None)
    monkeypatch.setitem(sys.modules, "dateutil.tz", None)

    # Reload module to force ImportErrors
    reload(vertical_clock)

    # Override datetime.
    # This is key for testing as we can fix time.
    monkeypatch.setattr("libqtile.widget.vertical_clock.datetime", MockDatetime)

    class TestVerticalClock(vertical_clock.VerticalClock):
        def __init__(self, **config):
            vertical_clock.VerticalClock.__init__(self, **config)
            self.name = "verticalclock"

        def info(self):
            info = vertical_clock.VerticalClock.info(self)
            info["text"] = "|".join(layout.text for layout in self.layouts)
            return info

    yield TestVerticalClock


@pytest.fixture(scope="function")
def vclock_manager(manager_nospawn, request, patched_clock):
    class VClockConfig(BareConfig):
        screens = [
            Screen(
                left=Bar(
                    [
                        patched_clock(
                            **getattr(request, "param", dict()),
                        )
                    ],
                    30,
                )
            )
        ]

    manager_nospawn.start(VClockConfig)

    yield manager_nospawn


def config(**kwargs):
    return pytest.mark.parametrize("vclock_manager", [kwargs], indirect=True)


def test_vclock_default(vclock_manager):
    assert vclock_manager.c.widget["verticalclock"].info()["text"] == "10|20"


@config(format=["%H", "%M", "-", "%d", "%m", "%Y"])
def test_vclock_extra_lines(vclock_manager):
    assert vclock_manager.c.widget["verticalclock"].info()["text"] == "10|20|-|01|01|2024"


@pytest.mark.parametrize(
    "vclock_manager",
    [
        dict(fontsize=[10]),  # too few
        dict(fontsize=[10, 20, 30, 40]),  # too many
        dict(foreground=["fff"]),  # too few
        dict(foreground=["fff"] * 4),  # too many
    ],
    indirect=True,
)
def test_vclock_invalid_configs(vclock_manager):
    assert vclock_manager.c.bar["left"].info()["widgets"][0]["name"] == "configerrorwidget"
